@@ -12,25 +12,28 @@ is called with the same input arguments.
 from __future__ import with_statement
 import os
 import time
+import pathlib
 import pydoc
 import re
 import functools
 import traceback
 import warnings
 import inspect
+import sys
 import weakref
 
+from tokenize import open as open_py_source
+
 # Local imports
 from . import hashing
 from .func_inspect import get_func_code, get_func_name, filter_args
 from .func_inspect import format_call
 from .func_inspect import format_signature
-from ._memory_helpers import open_py_source
 from .logger import Logger, format_time, pformat
-from ._compat import _basestring, PY3_OR_LATER
 from ._store_backends import StoreBackendBase, FileSystemStoreBackend
 
 
+
 FIRST_LINE_TEXT = "# first line:"
 
 # TODO: The following object should have a data store object as a sub
@@ -85,7 +88,7 @@ def register_store_backend(backend_name, backend):
         The name of a class that implements the StoreBackendBase interface.
 
     """
-    if not isinstance(backend_name, _basestring):
+    if not isinstance(backend_name, str):
         raise ValueError("Store backend name should be a string, "
                          "'{0}' given.".format(backend_name))
     if backend is None or not issubclass(backend, StoreBackendBase):
@@ -101,9 +104,12 @@ def _store_backend_factory(backend, location, verbose=0, backend_options=None):
     if backend_options is None:
         backend_options = {}
 
+    if isinstance(location, pathlib.Path):
+        location = str(location)
+
     if isinstance(location, StoreBackendBase):
         return location
-    elif isinstance(location, _basestring):
+    elif isinstance(location, str):
         obj = None
         location = os.path.expanduser(location)
         # The location is not a local file system, we look in the
@@ -124,6 +130,12 @@ def _store_backend_factory(backend, location, verbose=0, backend_options=None):
         obj.configure(location, verbose=verbose,
                       backend_options=backend_options)
         return obj
+    elif location is not None:
+        warnings.warn(
+            "Instanciating a backend using a {} as a location is not "
+            "supported by joblib. Returning None instead.".format(
+                location.__class__.__name__), UserWarning)
+
 
     return None
 
@@ -138,7 +150,7 @@ def _get_func_fullname(func):
 def _build_func_identifier(func):
     """Build a roughly unique identifier for the cached function."""
     parts = []
-    if isinstance(func, _basestring):
+    if isinstance(func, str):
         parts.append(func)
     else:
         parts.append(_get_func_fullname(func))
@@ -211,7 +223,7 @@ class MemorizedResult(Logger):
                  mmap_mode=None, verbose=0, timestamp=None, metadata=None):
         Logger.__init__(self)
         self.func_id = _build_func_identifier(func)
-        if isinstance(func, _basestring):
+        if isinstance(func, str):
             self.func = func
         else:
             self.func = self.func_id
@@ -247,8 +259,19 @@ class MemorizedResult(Logger):
                                    metadata=self.metadata)
         else:
             msg = None
-        return self.store_backend.load_item(
-            [self.func_id, self.args_id], msg=msg, verbose=self.verbose)
+
+        try:
+            return self.store_backend.load_item(
+                [self.func_id, self.args_id], msg=msg, verbose=self.verbose)
+        except ValueError as exc:
+            new_exc = KeyError(
+                "Error while trying to load a MemorizedResult's value. "
+                "It seems that this folder is corrupted : {}".format(
+                    os.path.join(
+                        self.store_backend.location, self.func_id,
+                        self.args_id)
+                ))
+            raise new_exc from exc
 
     def clear(self):
         """Clear value from cache"""
@@ -426,6 +449,9 @@ class MemorizedFunc(Logger):
             doc = func.__doc__
         self.__doc__ = 'Memoized version of %s' % doc
 
+        self._func_code_info = None
+        self._func_code_id = None
+
     def _cached_call(self, args, kwargs, shelving=False):
         """Call wrapped function and cache result, or read cache if available.
 
@@ -454,12 +480,16 @@ class MemorizedFunc(Logger):
         metadata: dict
             Some metadata about wrapped function call (see _persist_input()).
         """
-        # Compare the function code with the previous to see if the
-        # function code has changed
         func_id, args_id = self._get_output_identifiers(*args, **kwargs)
         metadata = None
         msg = None
+
+        # Wether or not the memorized function must be called
+        must_call = False
+
         # FIXME: The statements below should be try/excepted
+        # Compare the function code with the previous to see if the
+        # function code has changed
         if not (self._check_previous_func_code(stacklevel=4) and
                 self.store_backend.contains_item([func_id, args_id])):
             if self._verbose > 10:
@@ -469,16 +499,7 @@ class MemorizedFunc(Logger):
                           .format(name, args_id,
                                   self.store_backend.
                                   get_cached_func_info([func_id])['location']))
-            out, metadata = self.call(*args, **kwargs)
-            if self.mmap_mode is not None:
-                # Memmap the output at the first call to be consistent with
-                # later calls
-                if self._verbose:
-                    msg = _format_load_msg(func_id, args_id,
-                                           timestamp=self.timestamp,
-                                           metadata=metadata)
-                out = self.store_backend.load_item([func_id, args_id], msg=msg,
-                                                   verbose=self._verbose)
+            must_call = True
         else:
             try:
                 t0 = time.time()
@@ -507,11 +528,45 @@ class MemorizedFunc(Logger):
                 self.warn('Exception while loading results for '
                           '{}\n {}'.format(signature, traceback.format_exc()))
 
-                out, metadata = self.call(*args, **kwargs)
-                args_id = None
+                must_call = True
+
+        if must_call:
+            out, metadata = self.call(*args, **kwargs)
+            if self.mmap_mode is not None:
+                # Memmap the output at the first call to be consistent with
+                # later calls
+                if self._verbose:
+                    msg = _format_load_msg(func_id, args_id,
+                                           timestamp=self.timestamp,
+                                           metadata=metadata)
+                out = self.store_backend.load_item([func_id, args_id], msg=msg,
+                                                   verbose=self._verbose)
 
         return (out, args_id, metadata)
 
+    @property
+    def func_code_info(self):
+        # 3-tuple property containing: the function source code, source file,
+        # and first line of the code inside the source file
+        if hasattr(self.func, '__code__'):
+            if self._func_code_id is None:
+                self._func_code_id = id(self.func.__code__)
+            elif id(self.func.__code__) != self._func_code_id:
+                # Be robust to dynamic reassignments of self.func.__code__
+                self._func_code_info = None
+
+        if self._func_code_info is None:
+            # Cache the source code of self.func . Provided that get_func_code
+            # (which should be called once on self) gets called in the process
+            # in which self.func was defined, this caching mechanism prevents
+            # undesired cache clearing when the cached function is called in
+            # an environement where the introspection utilitiees get_func_code
+            # relies on do not work (typicially, in joblib child processes).
+            # See #1035 for  more info
+            # TODO (pierreglaser): do the same with get_func_name?
+            self._func_code_info = get_func_code(self.func)
+        return self._func_code_info
+
     def call_and_shelve(self, *args, **kwargs):
         """Call wrapped function, cache result and return a reference.
 
@@ -536,11 +591,19 @@ class MemorizedFunc(Logger):
         return self._cached_call(args, kwargs)[0]
 
     def __getstate__(self):
-        """ We don't store the timestamp when pickling, to avoid the hash
-            depending from it.
-        """
+        # Make sure self.func's source is introspected prior to being pickled -
+        # code introspection utilities typically do not work inside child
+        # processes
+        _ = self.func_code_info
+
+        # We don't store the timestamp when pickling, to avoid the hash
+        # depending from it.
         state = self.__dict__.copy()
         state['timestamp'] = None
+
+        # Invalidate the code id as id(obj) will be different in the child
+        state['_func_code_id'] = None
+
         return state
 
     # ------------------------------------------------------------------------
@@ -576,12 +639,8 @@ class MemorizedFunc(Logger):
 
         # Also store in the in-memory store of function hashes
         is_named_callable = False
-        if PY3_OR_LATER:
-            is_named_callable = (hasattr(self.func, '__name__') and
-                                 self.func.__name__ != '<lambda>')
-        else:
-            is_named_callable = (hasattr(self.func, 'func_name') and
-                                 self.func.func_name != '<lambda>')
+        is_named_callable = (hasattr(self.func, '__name__') and
+                             self.func.__name__ != '<lambda>')
         if is_named_callable:
             # Don't do this for lambda functions or strange callable
             # objects, as it ends up being too fragile
@@ -616,7 +675,7 @@ class MemorizedFunc(Logger):
         # Here, we go through some effort to be robust to dynamically
         # changing code and collision. We cannot inspect.getsource
         # because it is not reliable when using IPython's magic "%run".
-        func_code, source_file, first_line = get_func_code(self.func)
+        func_code, source_file, first_line = self.func_code_info
         func_id = _build_func_identifier(self.func)
 
         try:
@@ -688,7 +747,7 @@ class MemorizedFunc(Logger):
             self.warn("Clearing function cache identified by %s" % func_id)
         self.store_backend.clear_path([func_id, ])
 
-        func_code, _, first_line = get_func_code(self.func)
+        func_code, _, first_line = self.func_code_info
         self._write_func_code(func_code, first_line)
 
     def call(self, *args, **kwargs):
@@ -869,7 +928,7 @@ class Memory(Logger):
             location = cachedir
 
         self.location = location
-        if isinstance(location, _basestring):
+        if isinstance(location, str):
             location = os.path.join(location, 'joblib')
 
         self.store_backend = _store_backend_factory(
